from typing import TYPE_CHECKING

from sqlalchemy import Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_run import Run


class Script(Model):
    __tablename__ = 'script'
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    current: Mapped[bool] = mapped_column(default=False)
    script: Mapped[str] = mapped_column(Text)

    runs: Mapped[list['Run']] = relationship(back_populates='script')



from typing import cast

import strawberry
from sqlalchemy import select
from strawberry.types import Info

from nextline_rdb.db import DB
from nextline_rdb.models import Run


async def mutate_delete_runs(info: Info, ids: list[int]) -> list[int]:
    db = cast(DB, info.context['db'])
    async with db.session.begin() as session:
        stmt = select(Run).filter(Run.id.in_(ids))
        runs = (await session.execute(stmt)).scalars().all()
        # ic(runs)
        for run in runs:
            await session.delete(run)
        ret = [run.id for run in runs]
    return ret


@strawberry.type
class MutationRDB:
    delete_runs: list[int] = strawberry.field(resolver=mutate_delete_runs)


@strawberry.type
class Mutation:
    @strawberry.field
    async def rdb(self) -> MutationRDB:
        return MutationRDB()



from typing import Optional, Protocol, TypeVar

from graphql import GRAPHQL_MAX_INT, GRAPHQL_MIN_INT
from hypothesis import strategies as st

from nextline_rdb.utils import safe_max, safe_min

SQLITE_INT_MIN = -9_223_372_036_854_775_808  # 2 ** 63 * -1
SQLITE_INT_MAX = 9_223_372_036_854_775_807  # 2 ** 63 - 1

T = TypeVar('T')


class StMinMaxValuesFactory(Protocol[T]):
    def __call__(
        self, *, min_value: Optional[T] = None, max_value: Optional[T] = None
    ) -> st.SearchStrategy[T]:
        ...


def st_none_or(st_: st.SearchStrategy[T]) -> st.SearchStrategy[Optional[T]]:
    '''A strategy for `None` or values from another strategy.

    >>> v = st_none_or(st.integers()).example()
    >>> v is None or isinstance(v, int)
    True
    '''
    return st.one_of(st.none(), st_)


def st_sqlite_ints(
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
) -> st.SearchStrategy[int]:
    '''A strategy for integers within the range of SQLite's `INTEGER` type.

    >>> int_ = st_sqlite_ints().example()
    >>> int_ >= SQLITE_INT_MIN
    True

    >>> int_ <= SQLITE_INT_MAX
    True
    '''
    min_value = safe_max((min_value, SQLITE_INT_MIN))
    max_value = safe_min((max_value, SQLITE_INT_MAX))
    return st.integers(min_value=min_value, max_value=max_value)


def st_graphql_ints(
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
) -> st.SearchStrategy[int]:
    '''A strategy for integers within the range of GraphQL's `Int` type.

    Note: GraphQL's `Int` type is a 32-bit signed integer, i.e., -2,147,483,648
    to 2,147,483,647, while SQLite's `INTEGER` type is a 64-bit signed integer.
    The range of GraphQL's `Int` type is a subset of SQLite's `INTEGER` type.

    >>> int_ = st_graphql_ints().example()
    >>> int_ >= GRAPHQL_MIN_INT
    True

    >>> int_ <= GRAPHQL_MAX_INT
    True
    '''
    min_value = safe_max((min_value, GRAPHQL_MIN_INT))
    max_value = safe_min((max_value, GRAPHQL_MAX_INT))
    return st.integers(min_value=min_value, max_value=max_value)


def st_ranges(
    st_: StMinMaxValuesFactory[T],
    min_start: Optional[T] = None,
    max_start: Optional[T] = None,
    min_end: Optional[T] = None,
    max_end: Optional[T] = None,
    allow_start_none: bool = True,
    allow_end_none: bool = True,
    let_end_none_if_start_none: bool = False,
    allow_equal: bool = True,
) -> st.SearchStrategy[tuple[Optional[T], Optional[T]]]:
    '''Generate two values (start, end) from a strategy, where start <= end.

    The minimum and maximum values can be specified by `min_start`,
    `max_start`, `min_end`, `max_end`.

    `start` (`end`) can be `None` if `allow_start_none` (`allow_end_none`) is `True`.

    If `let_end_none_if_start_none` is `True`, `end` will be always `None` when
    `start` is `None` regardless of `allow_end_none`.

    If `allow_equal` is `False`, `start` and `end` cannot be equal, i.e., `start < end`.

    >>> start, end = st_ranges(
    ...     st.integers,
    ...     min_start=0,
    ...     max_end=10,
    ...     allow_start_none=False,
    ...     allow_end_none=False,
    ... ).example()

    The results can be, for example, used as min_value and max_value of st.integers().

    >>> i = st.integers(min_value=start, max_value=end).example()
    >>> start <= i <= end
    True

    The results can also be used as min_size and max_size of st.lists().

    >>> l = st.lists(st.integers(), min_size=start, max_size=end).example()
    >>> start <= len(l) <= end
    True

    '''

    def st_start() -> st.SearchStrategy[Optional[T]]:
        _max_start = safe_min((max_start, max_end))
        _st = st_(min_value=min_start, max_value=_max_start)
        return st_none_or(_st) if allow_start_none else _st

    def st_end(start: T | None) -> st.SearchStrategy[Optional[T]]:
        _min_end = safe_max((min_start, start, min_end))
        if min_end is not None and max_end is not None:
            assert min_end <= max_end  # type: ignore
        if start is None and let_end_none_if_start_none:
            return st.none()
        _st = st_(min_value=_min_end, max_value=max_end)
        if start is not None and not allow_equal:
            _st = _st.filter(lambda x: x > start)  # type: ignore
        return st_none_or(_st) if allow_end_none else _st

    return st_start().flatmap(
        lambda start: st.tuples(st.just(start), st_end(start=start))
    )



from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.utils.strategies import st_none_or

from ... import Model, Trace
from .. import st_model_run, st_model_trace_list


@given(st.data())
async def test_st_model_trace_lists(data: st.DataObject) -> None:
    run = data.draw(st_none_or(st_model_run(generate_traces=False)))
    max_size = data.draw(st.integers(min_value=0, max_value=3))
    traces = data.draw(st_model_trace_list(run=run, max_size=max_size))

    assert len(traces) <= max_size

    if traces:
        runs = set(trace.run for trace in traces)
        assert len(runs) == 1
        assert run is None or run is runs.pop()

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add_all(traces)

        async with db.session() as session:
            stmt = select(Trace)
            traces_ = (await session.scalars(stmt)).all()
            session.expunge_all()

    assert repr(traces) == repr(traces_)



import asyncio
from datetime import timezone
from logging import getLogger

from nextline.plugin.spec import Context, hookimpl
from nextline.spawned import RunArg
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB
from nextline_rdb.models import CurrentScript, Run, Script


class WriteRunTable:
    def __init__(self, db: DB) -> None:
        self._db = db
        self._logger = getLogger(__name__)

    @hookimpl
    async def on_start_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        assert (running_process := context.running_process)
        started_at = running_process.process_created_at
        assert started_at.tzinfo is timezone.utc
        started_at = started_at.replace(tzinfo=None)
        run_no = run_arg.run_no
        async with self._db.session.begin() as session:
            script = await self._find_script(run_arg, session)
            run = Run(
                run_no=run_no, state='running', started_at=started_at, script=script
            )
            session.add(run)

    async def _find_script(
        self, run_arg: RunArg, session: AsyncSession
    ) -> Script | None:
        statement = self._str_statement_or_none(run_arg)
        current_script = await self._load_current_script(session)
        match statement, current_script:
            case None, None:
                return None
            case None, CurrentScript() as cs:
                self._logger.warning(
                    'CurrentScript is not None, but run_arg.statement is None'
                )
                await session.delete(cs)
            case str(statement_), None:
                self._logger.warning(
                    'CurrentScript is None, but run_arg.statement is not None'
                )
                return Script(script=statement_, current=True)
            case str(statement_), CurrentScript() as cs:
                if not cs.script.script == statement:
                    self._logger.warning(
                        'The statement in CurrentScript is different from run_arg.statement'
                    )
                    s = Script(script=statement_)
                    cs.script = s
                return cs.script
        return None

    def _str_statement_or_none(self, run_arg: RunArg) -> str | None:
        if isinstance(run_arg.statement, str):
            return run_arg.statement
        return None

    async def _load_current_script(self, session: AsyncSession) -> CurrentScript | None:
        stmt = select(CurrentScript).options(selectinload(CurrentScript.script))
        return (await session.execute(stmt)).scalar_one_or_none()

    @hookimpl
    async def on_end_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        assert (exited_process := context.exited_process)
        assert (returned := exited_process.returned)
        ended_at = exited_process.process_exited_at
        assert ended_at.tzinfo is timezone.utc
        ended_at = ended_at.replace(tzinfo=None)
        run_no = run_arg.run_no
        async with self._db.session.begin() as session:
            stmt = select(Run).filter_by(run_no=run_no)
            while not (run := (await session.execute(stmt)).scalar_one_or_none()):
                await asyncio.sleep(0)
            run.state = 'finished'
            run.ended_at = ended_at
            run.exception = returned.fmt_exc



import datetime as dt
from typing import Optional

from hypothesis import strategies as st


def st_datetimes(
    min_value: Optional[dt.datetime] = None,
    max_value: Optional[dt.datetime] = None,
) -> st.SearchStrategy[dt.datetime]:
    '''A strategy for naive `datetime` objects without imaginary datetimes or folds.

    Note: timezones and folds are not supported by SQLite.

    >>> dt_ = st_datetimes().example()
    >>> dt_.tzinfo is None
    True

    >>> dt_.fold
    0
    '''
    if min_value is None:
        min_value = dt.datetime.min

    if max_value is None:
        max_value = dt.datetime.max

    return st.datetimes(
        min_value=min_value,
        max_value=max_value,
        allow_imaginary=False,
    ).map(lambda dt_: dt_.replace(fold=0))



from pathlib import Path

import pytest
from alembic.config import Config

from nextline_rdb.db import ALEMBIC_INI


@pytest.fixture
def alembic_config() -> Config:
    assert Path(ALEMBIC_INI).is_file()
    config = Config(ALEMBIC_INI)
    return config


@pytest.fixture
def alembic_config_in_memory(alembic_config: Config) -> Config:
    config = alembic_config
    url = 'sqlite+aiosqlite://'
    config.set_main_option('sqlalchemy.url', url)
    return config


@pytest.fixture
def alembic_config_temp_sqlite(
    alembic_config: Config, tmp_path_factory: pytest.TempPathFactory
) -> Config:
    config = alembic_config
    dir = tmp_path_factory.mktemp('db')
    url = f'sqlite+aiosqlite:///{dir}/db.sqlite'
    config.set_main_option('sqlalchemy.url', url)
    return config



from sqlalchemy import ForeignKey, event
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    Session,
    mapped_column,
    relationship,
    sessionmaker,
)

from nextline_rdb.models import ReprMixin


class Model(ReprMixin, DeclarativeBase):
    __abstract__ = True
    id: Mapped[int] = mapped_column(primary_key=True, index=True)


class Foo(Model):
    __tablename__ = 'foo'
    bars: Mapped[list['Bar']] = relationship(back_populates='foo')


class Bar(Model):
    __tablename__ = 'bar'
    foo_id: Mapped[int] = mapped_column(ForeignKey('foo.id'))
    foo: Mapped[Foo] = relationship(back_populates='bars')


def register_session_events(session: sessionmaker[Session]) -> None:
    new: set[Model]

    @event.listens_for(session, 'before_flush')
    def _before_flush(session: Session, flush_context, instances):
        nonlocal new
        new = set(session.new)

    @event.listens_for(session, 'after_flush_postexec')
    def _after_flush_postexec(session: Session, flush_context):
        pass



def ensure_async_url(url: str) -> str:
    '''Convert an SQLAlchemy URL to an async one if necessary.

    Support only SQLite for now.

    >>> ensure_async_url('sqlite://')
    'sqlite+aiosqlite://'

    >>> ensure_async_url('sqlite+aiosqlite://')
    'sqlite+aiosqlite://'

    >>> ensure_async_url('sqlite:///foo.db')
    'sqlite+aiosqlite:///foo.db'

    >>> ensure_async_url('sqlite+aiosqlite:///foo.db')
    'sqlite+aiosqlite:///foo.db'

    >>> ensure_async_url('postgresql://scott:tiger@localhost/database')
    'postgresql://scott:tiger@localhost/database'

    '''
    scheme, rest = url.split("://", 1)
    if 'sqlite' not in scheme:
        return url
    if 'aiosqlite' in scheme:
        return url
    scheme = scheme.replace('sqlite', 'sqlite+aiosqlite')
    return f'{scheme}://{rest}'


def ensure_sync_url(url: str) -> str:
    '''Convert an SQLAlchemy URL to a sync one if necessary.

    Support only SQLite for now.

    >>> ensure_sync_url('sqlite://')
    'sqlite://'

    >>> ensure_sync_url('sqlite+aiosqlite://')
    'sqlite://'

    >>> ensure_sync_url('sqlite:///foo.db')
    'sqlite:///foo.db'

    >>> ensure_sync_url('sqlite+aiosqlite:///foo.db')
    'sqlite:///foo.db'

    >>> ensure_sync_url('postgresql+asyncpg://scott:tiger@localhost/database')
    'postgresql+asyncpg://scott:tiger@localhost/database'

    '''
    scheme, rest = url.split("://", 1)
    if 'sqlite' not in scheme:
        return url
    if 'aiosqlite' not in scheme:
        return url
    scheme = scheme.replace('sqlite+aiosqlite', 'sqlite')
    return f'{scheme}://{rest}'



import base64
from collections.abc import Callable
from functools import partial
from typing import Optional, Type, TypeVar

from sqlalchemy import func, inspect, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.sql.selectable import Select

from nextline_rdb.pagination import Sort, load_models

from .connection import Connection, Edge, query_connection

_M = TypeVar('_M', bound=DeclarativeBase)  # Model
_N = TypeVar('_N')  # Node


def encode_id(id: int) -> str:
    return base64.b64encode(f'{id}'.encode()).decode()


def decode_id(cursor: str) -> int:
    return int(base64.b64decode(cursor).decode())


async def load_connection(
    session: AsyncSession,
    Model: Type[_M],
    create_node_from_model: Callable[[_M], _N],
    *,
    select_model: Optional[Select[tuple[_M]]] = None,
    sort: Optional[Sort] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[_N]:
    query_edges = partial(
        load_edges,
        session=session,
        Model=Model,
        create_node_from_model=create_node_from_model,
        select_model=select_model,
        sort=sort,
    )

    query_total_count = partial(
        load_total_count,
        session=session,
        Model=Model,
        select_model=select_model,
    )

    return await query_connection(
        query_edges,
        query_total_count,
        before,
        after,
        first,
        last,
    )


async def load_total_count(
    session: AsyncSession,
    Model: Type[_M],
    *,
    select_model: Optional[Select[tuple[_M]]] = None,
) -> int:
    if select_model is None:
        select_model = select(Model)
    cte = select_model.cte()
    stmt = select(func.count()).select_from(cte)
    total_count = (await session.execute(stmt)).scalar() or 0
    return total_count


async def load_edges(
    session: AsyncSession,
    Model: Type[_M],
    create_node_from_model: Callable[[_M], _N],
    *,
    select_model: Optional[Select[tuple[_M]]] = None,
    sort: Optional[Sort] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> list[Edge[_N]]:
    # TODO: handle multiple primary keys
    primary_keys = list(inspect(Model).primary_key)
    assert len(primary_keys) == 1, 'Multiple primary keys are not supported'
    id_field = primary_keys[0].name

    models = await load_models(
        session,
        Model,
        id_field,
        select_model=select_model,
        sort=sort,
        before=before if before is None else decode_id(before),
        after=after if after is None else decode_id(after),
        first=first,
        last=last,
    )

    nodes = [create_node_from_model(m) for m in models]

    edges = [Edge(node=n, cursor=encode_id(getattr(n, id_field))) for n in nodes]

    return edges



from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB

from ... import Model, Prompt
from .. import st_model_prompt


@given(st.data())
async def test_st_model_prompt(data: st.DataObject) -> None:
    prompt = data.draw(st_model_prompt())

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(prompt)
        async with db.session() as session:
            select_prompt = select(Prompt)
            prompt_ = await session.scalar(
                select_prompt.options(
                    selectinload(Prompt.run), selectinload(Prompt.trace)
                )
            )
            assert prompt_
            session.expunge_all()
    assert repr(prompt) == repr(prompt_)



from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from nextline_rdb.models import NAMING_CONVENTION, ReprMixin

metadata = MetaData(naming_convention=dict(NAMING_CONVENTION))


class Base(ReprMixin, DeclarativeBase):
    metadata = metadata


class Entity(Base):
    __tablename__ = 'entity'
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    num: Mapped[int | None]
    txt: Mapped[str | None]



from datetime import datetime
from typing import TYPE_CHECKING

from sqlalchemy import ForeignKey, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_run import Run
    from .model_trace import Trace


class Prompt(Model):
    __tablename__ = "prompt"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int]
    trace_no: Mapped[int]
    prompt_no: Mapped[int]
    open: Mapped[bool]
    event: Mapped[str]
    started_at: Mapped[datetime]
    file_name: Mapped[str | None]
    line_no: Mapped[int | None]
    stdout: Mapped[str | None]
    command: Mapped[str | None]
    ended_at: Mapped[datetime | None]

    run_id: Mapped[int] = mapped_column(ForeignKey('run.id'))
    run: Mapped['Run'] = relationship(back_populates='prompts')

    trace_id: Mapped[int] = mapped_column(ForeignKey('trace.id'))
    trace: Mapped['Trace'] = relationship(back_populates='prompts')

    __table_args__ = (UniqueConstraint("run_no", "prompt_no"),)



import datetime
from typing import TYPE_CHECKING, Annotated, Optional

import strawberry

from nextline_rdb import models as db_models

if TYPE_CHECKING:
    from .run_node import RunNode
    from .trace_node import TraceNode


@strawberry.type
class StdoutNode:
    _model: strawberry.Private[db_models.Stdout]
    id: int
    run_no: int
    trace_no: int
    text: Optional[str] = None
    written_at: Optional[datetime.datetime] = None

    @strawberry.field
    def run(self) -> Annotated['RunNode', strawberry.lazy('.run_node')]:
        from .run_node import RunNode

        return RunNode.from_model(self._model.run)

    @strawberry.field
    def trace(self) -> Annotated['TraceNode', strawberry.lazy('.trace_node')]:
        from .trace_node import TraceNode

        return TraceNode.from_model(self._model.trace)

    @classmethod
    def from_model(cls: type['StdoutNode'], model: db_models.Stdout):
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            trace_no=model.trace_no,
            text=model.text,
            written_at=model.written_at,
        )



from threading import Thread


def run():
    t1 = Thread(target=f1)
    t1.start()
    t2 = Thread(target=f2)
    t2.start()
    t2.join()
    t1.join()
    return


def f1():
    for _ in range(5):
        pass
    return


def f2():
    for _ in range(5):
        pass
    return



from collections.abc import Callable
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import Optional, Type

from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy import Connection, Engine, MetaData, event
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Session, sessionmaker

import nextline_rdb
from nextline_rdb import models
from nextline_rdb.utils import ensure_async_url

_PACKAGE_TOP = Path(nextline_rdb.__file__).resolve().parent
ALEMBIC_INI = str(_PACKAGE_TOP / 'alembic' / 'alembic.ini')


assert Path(ALEMBIC_INI).is_file()


class DB:
    '''The interface to the async SQLAlchemy database.

    >>> async def main():
    ...     # An example usage
    ...     async with DB() as db:
    ...         # nested session contexts
    ...         async with db.session() as session:
    ...             async with session.begin():
    ...                 # commit automatically
    ...                 pass
    ...         # can begin directly
    ...         async with db.session.begin() as session:
    ...             # commit automatically
    ...             pass
    ...
    ...     # An alternative usage
    ...     db = DB()
    ...     await db.start()
    ...     async with contextlib.aclosing(db):
    ...         async with db.session() as session:
    ...             async with session.begin():
    ...                 pass
    ...
    >>> import asyncio
    >>> import contextlib
    >>> asyncio.run(main())

    '''

    def __init__(
        self,
        url: Optional[str] = None,
        create_async_engine_kwargs: Optional[dict] = None,  # e.g., {'echo': True}
        model_base_class: Type[DeclarativeBase] = models.Model,
        use_migration: bool = True,
        migration_revision_target: str = 'head',
        alembic_ini_path: str | PathLike = ALEMBIC_INI,
        register_session_events: Optional[
            Callable[[sessionmaker[Session]], None]
        ] = None,
    ):
        url = url or 'sqlite+aiosqlite://'
        self.url = ensure_async_url(url)
        self.create_async_engine_kwargs = create_async_engine_kwargs or {}
        self.model_base_class = model_base_class
        self.metadata = self.model_base_class.metadata
        self.use_migration = use_migration
        self.migration_revision_target = migration_revision_target
        self.alembic_ini_path = alembic_ini_path

        self.engine = create_async_engine(self.url, **self.create_async_engine_kwargs)
        self.migration_revision: str | None = None

        self._logger = getLogger(__name__)
        self._logger.info(f'Async SQLAlchemy DB URL: {self.url}')

        self._register_session_events = register_session_events

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.url!r}>'

    async def start(self) -> None:
        if self.use_migration:
            await self._migrate()
        else:
            await self._define_tables()

        self.migration_revision = await self._get_current_revision()
        self._logger.info(f'Alembic migration version: {self.migration_revision!s}')

        # As of SQLAlchemy 2.0, the async session maker cannot be use for
        # registering event hooks with the event.listen_for() function. Use the
        # sync session maker instead as described in the SQLAlchemy documentation:
        # https://docs.sqlalchemy.org/en/20/orm/extensions/asyncio.html#examples-of-event-listeners-with-async-engines-sessions-sessionmakers

        sync_maker = sessionmaker()
        self.session = async_sessionmaker(
            self.engine, expire_on_commit=False, sync_session_class=sync_maker
        )

        if self._register_session_events:
            self._register_session_events(sync_maker)

    async def _migrate(self) -> None:
        assert Path(self.alembic_ini_path).is_file()
        config = Config(self.alembic_ini_path)
        await migrate(
            engine=self.engine,
            metadata=self.metadata,
            config=config,
            target=self.migration_revision_target,
        )

    async def _get_current_revision(self) -> str | None:
        def _fn(conn: Connection) -> str | None:
            context = MigrationContext.configure(conn)
            return context.get_current_revision()

        async with self.engine.connect() as conn:
            return await conn.run_sync(_fn)

    async def _define_tables(self) -> None:
        async with self.engine.begin() as conn:
            await conn.run_sync(self.metadata.create_all)

    async def aclose(self) -> None:
        await self.engine.dispose()

    async def __aenter__(self) -> 'DB':
        await self.start()
        return self

    async def __aexit__(self, *_, **__) -> None:
        await self.aclose()


async def migrate(
    engine: AsyncEngine,
    metadata: MetaData,
    config: Config,
    target: str = 'head',
) -> None:
    script = ScriptDirectory.from_config(config)

    def upgrade(rev: str, context):  # type: ignore[no-untyped-def]
        del context
        return script._upgrade_revs(target, rev)

    context = EnvironmentContext(config, script, fn=upgrade)

    def do_run_migrations(connection: Connection) -> None:
        context.configure(
            connection=connection,
            target_metadata=metadata,
            render_as_batch=True,
        )
        with context.begin_transaction():
            context.run_migrations()

    async with engine.connect() as connection:
        await connection.run_sync(do_run_migrations)


@event.listens_for(Engine, 'connect')
def set_sqlite_pragma(dbapi_connection, connection_record):
    '''Enable foreign key constraints in SQLite.

    The code copied from the SQLAlchemy documentation:
    https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support
    '''
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()



"""relate stdout to trace

Revision ID: 5a08750d6760
Revises: de90b787e92f
Create Date: 2022-03-16 19:01:00.628340

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '5a08750d6760'
down_revision = 'de90b787e92f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('stdout', schema=None) as batch_op:
        batch_op.add_column(sa.Column('trace_no', sa.Integer(), nullable=False))
        batch_op.add_column(sa.Column('trace_id', sa.Integer(), nullable=False))
        batch_op.create_foreign_key(
            batch_op.f('fk_stdout_trace_id_trace'), 'trace', ['trace_id'], ['id']
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('stdout', schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f('fk_stdout_trace_id_trace'), type_='foreignkey'
        )
        batch_op.drop_column('trace_id')
        batch_op.drop_column('trace_no')

    # ### end Alembic commands ###



from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.utils import safe_compare as sc
from nextline_rdb.utils.strategies import st_datetimes, st_ranges

from .. import st_started_at_ended_at


@given(st.data())
def test_st_started_at_ended_at(data: st.DataObject) -> None:
    min_start, max_start = data.draw(st_ranges(st_=st_datetimes))
    min_end, max_end = data.draw(st_ranges(st_=st_datetimes, min_start=min_start))
    allow_start_none = data.draw(st.booleans())
    allow_end_none = data.draw(st.booleans())
    allow_equal = data.draw(st.booleans())

    start, end = data.draw(
        st_started_at_ended_at(
            min_start=min_start,
            max_start=max_start,
            min_end=min_end,
            max_end=max_end,
            allow_start_none=allow_start_none,
            allow_end_none=allow_end_none,
            allow_equal=allow_equal,
        )
    )

    if not start:
        assert not end

    if not allow_start_none:
        assert start is not None

    if start is not None and not allow_end_none:
        assert end is not None

    if allow_equal:
        assert sc(start) <= sc(end)
    else:
        assert sc(start) < sc(end)

    assert sc(min_start) <= sc(start) <= sc(max_start)
    assert sc(min_end) <= sc(end) <= sc(max_end)



from typing import Any, TypeVar

from sqlalchemy import inspect, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase

T = TypeVar('T', bound=DeclarativeBase)


async def load_all(session: AsyncSession, model_base_class: type[T]) -> list[T]:
    '''All rows of all tables in the database.

    Sorted by the ORM class names and primary keys.
    '''
    objs = [
        m
        for c in all_declared_models_based_on(model_base_class)
        for m in (await session.scalars(select(c))).all()
    ]
    objs = sorted(objs, key=class_name_and_primary_keys_of)
    return objs


def all_declared_models_based_on(model_base_class: type[T]) -> list[type[T]]:
    '''The ORM classes inheriting from the base class.'''
    return [m.class_ for m in model_base_class.registry.mappers]


def class_name_and_primary_keys_of(instance: DeclarativeBase) -> tuple[Any, ...]:
    '''A tuple of the ORM class name followed by the primary key values.'''
    class_name = type(instance).__name__
    vals = primary_keys_of(instance)
    return (class_name,) + vals


def primary_keys_of(instance: DeclarativeBase) -> tuple[Any, ...]:
    '''A tuple of the primary key values.'''
    cls = type(instance)
    instance_mapper = inspect(cls)
    column_names = [attr.key for attr in instance_mapper.primary_key]
    values = tuple(getattr(instance, name) for name in column_names)
    return values



from collections.abc import Iterable
from typing import Any, Optional, TypeVar

T = TypeVar('T')


def safe_min(vals: Iterable[T], default: Optional[T] = None) -> Optional[T]:
    '''The smallest item in `vals` that is not `None`.

    >>> safe_min([None, 1, 2, None])
    1

    It returns `None` if `vals` is empty or all items in `vals` are `None`.
    >>> print(safe_min([None, None]))
    None

    >>> print(safe_min([]))
    None

    If `default` is given, it returns `default` instead of `None`.
    >>> safe_min([None, None], default=-1)
    -1

    >>> safe_min([], default=-1)
    -1
    '''
    return min((v for v in vals if v is not None), default=default)  # type: ignore


def safe_max(vals: Iterable[T], default: Optional[T] = None) -> Optional[T]:
    '''The largest item in `vals` that is not `None`.

    >>> safe_max([None, 1, 2, None])
    2

    It returns `None` if `vals` is empty or all items in `vals` are `None`.
    >>> print(safe_max([None, None]))
    None

    >>> print(safe_max([]))
    None

    If `default` is given, it returns `default` instead of `None`.
    >>> safe_max([None, None], default=-1)
    -1

    >>> safe_max([], default=-1)
    -1
    '''
    return max((v for v in vals if v is not None), default=default)  # type: ignore


class GreaterAndLessThanAny:
    '''True for all inequality comparisons.

    >>> GreaterAndLessThanAny() < 1
    True

    >>> GreaterAndLessThanAny() > 1
    True

    >>> GreaterAndLessThanAny() <= 1
    True

    >>> GreaterAndLessThanAny() >= 1
    True

    '''

    def __le__(self, _: Any) -> bool:
        return True

    def __lt__(self, _: Any) -> bool:
        return True

    def __ge__(self, _: Any) -> bool:
        return True

    def __gt__(self, _: Any) -> bool:
        return True

    # def __eq__(self, _: Any) -> bool:
    #     return True

    def __repr__(self) -> str:
        return 'GreaterAndLessThanAny()'


def safe_compare(value: T | None) -> T | GreaterAndLessThanAny:
    '''The `value` itself if it is not `None`. Otherwise, an object that returns True for all inequality comparisons.

    This function helps you concisely write assertions that compare values that
    may be `None`.

    Suppose you have `min_` and `max_` that may be `None`
    >>> import random
    >>> min_ = random.choice([None, 1])
    >>> max_ = random.choice([None, 3])

    and `val` that should be in the range `[min_, max_]`:
    >>> val = 2

    Without this function, you need to check if `min_` and `max_` are `None`.

    >>> if min_ is not None:
    ...   assert min_ <= val

    >>> if max_ is not None:
    ...    assert val <= max_

    This function lets you write the same assertion in one line:

    >>> assert safe_compare(min_) <= val <= safe_compare(max_)

    '''
    if value is None:
        return GreaterAndLessThanAny()
    return value



"""add the table "current_script"

Revision ID: 245c4152e8d7
Revises: c72fa3ee6a1a
Create Date: 2024-02-02 15:38:18.902883

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '245c4152e8d7'
down_revision = 'c72fa3ee6a1a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    op.create_table(
        'current_script',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('script_id', sa.Integer(), nullable=False),
        sa.CheckConstraint('id = 1', name=op.f('ck_current_script_')),
        sa.ForeignKeyConstraint(
            ['script_id'],
            ['script.id'],
            name=op.f('fk_current_script_script_id_script'),
        ),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_current_script')),
    )
    with op.batch_alter_table('current_script', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_current_script_id'), ['id'], unique=False)
    op.execute('PRAGMA foreign_keys=ON;')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('current_script', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_current_script_id'))

    op.drop_table('current_script')
    op.execute('PRAGMA foreign_keys=ON;')
    # ### end Alembic commands ###



from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.db import DB
from nextline_rdb.utils import class_name_and_primary_keys_of, load_all
from nextline_rdb.utils.strategies import st_none_or, st_ranges

from ... import Model, Run
from .. import st_model_run_list, st_model_script_list


@given(st.data())
async def test_options(data: st.DataObject) -> None:
    # Generate options of the strategy to be tested
    generate_traces = data.draw(st.booleans())
    min_size, max_size = data.draw(
        st_ranges(
            st_=st.integers,
            min_start=0,
            max_end=4,
            allow_start_none=False,
            allow_end_none=False,
        )
    )
    assert isinstance(min_size, int)
    assert isinstance(max_size, int)

    scripts = data.draw(st_none_or(st_model_script_list(max_size=max_size)))

    # Call the strategy to be tested
    runs = data.draw(
        st_model_run_list(
            generate_traces=generate_traces,
            min_size=min_size,
            max_size=max_size,
            scripts=scripts,
        )
    )

    # Assert the generated values
    assert min_size <= len(runs) <= max_size
    run_nos = [run.run_no for run in runs]
    assert run_nos == sorted(run_nos)

    traces = [trace for run in runs for trace in run.traces]
    prompts = [prompt for run in runs for prompt in run.prompts]
    stdouts = [stdout for run in runs for stdout in run.stdouts]

    if generate_traces is False:
        assert not traces
        assert not prompts
        assert not stdouts

    started_ats = [run.started_at for run in runs if run.started_at]
    assert started_ats == sorted(started_ats)

    scripts_ = {run.script for run in runs if run.script}
    current_script = {script for script in scripts_ if script.current}
    assert len(current_script) <= 1
    if scripts is not None:
        assert scripts_ <= set(scripts)


@given(runs=st_model_run_list(generate_traces=True, min_size=0, max_size=3))
async def test_db(runs: list[Run]) -> None:
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add_all(runs)
            added = list(session.new)
            assert set(runs) <= set(added)
        added = sorted(added, key=class_name_and_primary_keys_of)
        repr_added = [repr(m) for m in added]

        async with db.session() as session:
            loaded = await load_all(session, Model)
            repr_loaded = [repr(m) for m in loaded]

    assert repr_added == repr_loaded



"""ensure at most one current script

Revision ID: c72fa3ee6a1a
Revises: f9a742bb2297
Create Date: 2024-02-02 14:02:52.427179

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm.session import Session

from nextline_rdb.alembic.models.rev_6e3cf7d9b6bf import Script
from nextline_rdb.utils import mark_last

# revision identifiers, used by Alembic.
revision = 'c72fa3ee6a1a'
down_revision = 'f9a742bb2297'
branch_labels = None
depends_on = None


def upgrade():
    with Session(bind=op.get_bind()) as session:
        with session.begin():
            select_scripts = (
                sa.select(Script).where(Script.current.is_(True)).order_by(Script.id)
            )
            scripts = session.execute(select_scripts).scalars().all()
            if len(scripts) > 1:
                for last, script in mark_last(scripts):
                    if not last:
                        script.current = False


def downgrade():
    pass



from typing import TypeVar

from graphql import GRAPHQL_MAX_INT, GRAPHQL_MIN_INT
from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.utils import safe_compare
from nextline_rdb.utils.strategies import (
    SQLITE_INT_MAX,
    SQLITE_INT_MIN,
    st_graphql_ints,
    st_ranges,
    st_sqlite_ints,
)

T = TypeVar('T')


@given(st.data())
def test_st_sqlite_ints(data: st.DataObject) -> None:
    min_, max_ = data.draw(
        st_ranges(st_=st.integers, max_start=SQLITE_INT_MAX, min_end=SQLITE_INT_MIN)
    )

    i = data.draw(st_sqlite_ints(min_value=min_, max_value=max_))

    assert SQLITE_INT_MIN <= i <= SQLITE_INT_MAX
    assert safe_compare(min_) <= i <= safe_compare(max_)


@given(st.data())
def test_st_graphql_ints(data: st.DataObject) -> None:
    min_, max_ = data.draw(
        st_ranges(st_=st.integers, max_start=GRAPHQL_MAX_INT, min_end=GRAPHQL_MIN_INT)
    )

    i = data.draw(st_graphql_ints(min_value=min_, max_value=max_))

    assert GRAPHQL_MIN_INT <= i <= GRAPHQL_MAX_INT
    assert safe_compare(min_) <= i <= safe_compare(max_)



__all__ = [
    'PromptNode',
    'RunNode',
    'StdoutNode',
    'TraceNode',
]

from .prompt_node import PromptNode
from .run_node import RunNode
from .stdout_node import StdoutNode
from .trace_node import TraceNode



from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.utils.strategies import st_ranges

from ... import Model, Prompt, Run, Stdout, Trace
from .. import st_model_run_list


@given(st.data())
async def test_options(data: st.DataObject) -> None:
    generate_traces = data.draw(st.booleans())
    min_size, max_size = data.draw(
        st_ranges(
            st_=st.integers,
            min_start=0,
            max_end=4,
            allow_start_none=False,
            allow_end_none=False,
        )
    )
    assert isinstance(min_size, int)
    assert isinstance(max_size, int)
    runs = data.draw(
        st_model_run_list(
            generate_traces=generate_traces, min_size=min_size, max_size=max_size
        )
    )
    assert min_size <= len(runs) <= max_size
    run_nos = [run.run_no for run in runs]
    assert run_nos == sorted(run_nos)

    traces = [trace for run in runs for trace in run.traces]
    prompts = [prompt for run in runs for prompt in run.prompts]
    stdouts = [stdout for run in runs for stdout in run.stdouts]

    if generate_traces is False:
        assert not traces
        assert not prompts
        assert not stdouts

    started_ats = [run.started_at for run in runs if run.started_at]
    assert started_ats == sorted(started_ats)


@given(runs=st_model_run_list(generate_traces=True, min_size=0, max_size=3))
async def test_db(runs: list[Run]) -> None:
    traces = [trace for run in runs for trace in run.traces]
    prompts = [prompt for run in runs for prompt in run.prompts]
    stdouts = [stdout for run in runs for stdout in run.stdouts]

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add_all(runs)

        async with db.session() as session:
            select_run = select(Run)
            runs_ = (await session.scalars(select_run)).all()
            select_trace = select(Trace)
            traces_ = (await session.scalars(select_trace)).all()
            select_prompt = select(Prompt)
            prompts_ = (await session.scalars(select_prompt)).all()
            select_stdout = select(Stdout)
            stdouts_ = (await session.scalars(select_stdout)).all()
            session.expunge_all()

    assert repr(runs) == repr(runs_)
    traces = sorted(traces, key=lambda m: m.id)
    assert repr(traces) == repr(sorted(traces_, key=lambda m: m.id))
    prompts = sorted(prompts, key=lambda m: m.id)
    assert repr(prompts) == repr(sorted(prompts_, key=lambda m: m.id))
    stdouts = sorted(stdouts, key=lambda m: m.id)
    assert repr(stdouts) == repr(sorted(stdouts_, key=lambda m: m.id))



"""delete "script" from the table "run"

Revision ID: f9a742bb2297
Revises: cafceacada62
Create Date: 2024-01-30 13:48:55.138737

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'f9a742bb2297'
down_revision = 'cafceacada62'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('run', schema=None) as batch_op:
        batch_op.drop_column('script_old')
    op.execute('PRAGMA foreign_keys=ON;')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('run', schema=None) as batch_op:
        batch_op.add_column(sa.Column('script_old', sa.TEXT(), nullable=True))
    op.execute('PRAGMA foreign_keys=ON;')

    # ### end Alembic commands ###



from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB

from ... import Model, Stdout
from .. import st_model_stdout


@given(st.data())
async def test_st_model_stdout(data: st.DataObject) -> None:
    stdout = data.draw(st_model_stdout())

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(stdout)
        async with db.session() as session:
            select_prompt = select(Stdout)
            stdout_ = await session.scalar(
                select_prompt.options(
                    selectinload(Stdout.run), selectinload(Stdout.trace)
                )
            )
            assert stdout_
            session.expunge_all()
    assert repr(stdout) == repr(stdout_)



from collections.abc import Iterable, Iterator
from typing import Tuple, TypeVar

_T = TypeVar('_T')


def mark_last(it: Iterable[_T]) -> Iterator[Tuple[bool, _T]]:
    '''Indicate the end of the iteration.

    >>> list(mark_last([1, 2, 3]))
    [(False, 1), (False, 2), (True, 3)]

    >>> list(mark_last([1]))
    [(True, 1)]

    >>> list(mark_last([]))
    []


    This function is copied from `c5backup`:
    https://github.com/simonsobs/c5backup/blob/v0.3.0/src/c5backup/util/gen.py

    '''
    first = True
    for i in it:
        if first:
            first = False
            p = i
            continue
        yield False, p
        p = i
    else:
        if not first:
            yield True, p



from typing import Optional

from hypothesis import strategies as st

from nextline_rdb.utils.strategies import st_python_scripts, st_none_or

from .. import Script


def st_model_script(current: Optional[bool] = None) -> st.SearchStrategy[Script]:
    st_current = st.just(current) if current is not None else st.booleans()
    return st.builds(Script, current=st_current, script=st_python_scripts())


@st.composite
def st_model_script_list(
    draw: st.DrawFn,
    min_size: int = 0,
    max_size: Optional[int] = None,
) -> list[Script]:
    scripts = draw(
        st.lists(st_model_script(current=False), min_size=min_size, max_size=max_size)
    )
    if scripts:
        current = draw(st_none_or(st.sampled_from(scripts)))
        if current is not None:
            current.current = True
    return scripts



import asyncio
from datetime import timezone

from nextline.events import OnEndTrace, OnStartTrace
from nextline.plugin.spec import Context, hookimpl
from nextline.types import TraceNo
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.models import Run, Trace


class WriteTraceTable:
    def __init__(self, db: DB) -> None:
        self._db = db
        self._running_trace_nos = set[TraceNo]()

    @hookimpl
    async def on_start_trace(self, event: OnStartTrace) -> None:
        async with self._db.session.begin() as session:
            stmt = select(Run).filter_by(run_no=event.run_no)
            while not (run := (await session.execute(stmt)).scalar_one_or_none()):
                await asyncio.sleep(0)
            trace = Trace(
                run_no=event.run_no,
                trace_no=event.trace_no,
                state='running',
                thread_no=event.thread_no,
                task_no=event.task_no,
                started_at=event.started_at,
                run=run,
            )
            session.add(trace)
        self._running_trace_nos.add(event.trace_no)

    @hookimpl
    async def on_end_trace(self, event: OnEndTrace) -> None:
        async with self._db.session.begin() as session:
            stmt = select(Trace).filter_by(run_no=event.run_no, trace_no=event.trace_no)
            while not (trace := (await session.execute(stmt)).scalar_one_or_none()):
                await asyncio.sleep(0)
            trace.state = 'finished'
            trace.ended_at = event.ended_at
        self._running_trace_nos.remove(event.trace_no)

    @hookimpl
    async def on_start_run(self) -> None:
        self._running_trace_nos.clear()

    @hookimpl
    async def on_end_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        assert (exited_process := context.exited_process)
        ended_at = exited_process.process_exited_at
        assert ended_at.tzinfo is timezone.utc
        ended_at = ended_at.replace(tzinfo=None)
        run_no = run_arg.run_no
        async with self._db.session.begin() as session:
            stmt = (
                select(Trace)
                .filter_by(run_no=run_no)
                .filter(Trace.trace_no.in_(self._running_trace_nos))
            )
            traces = (await session.execute(stmt)).scalars().all()
            for trace in traces:
                trace.state = 'finished'
                trace.ended_at = ended_at
        self._running_trace_nos.clear()



from nextline.plugin.spec import Context, hookimpl
from nextline.spawned import RunArg
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB
from nextline_rdb.models import CurrentScript, Script


class WriteScriptTable:
    def __init__(self, db: DB) -> None:
        self._db = db

    @hookimpl
    async def on_initialize_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        statement = self._str_statement_or_none(run_arg)
        if statement is not None:
            await self._on_initialize_run_with_statement(statement)
        else:
            await self._on_initialize_run_without_statement()

    def _str_statement_or_none(self, run_arg: RunArg) -> str | None:
        if isinstance(run_arg.statement, str):
            return run_arg.statement
        return None

    async def _on_initialize_run_without_statement(self) -> None:
        async with self._db.session.begin() as session:
            current_script = await self._load_current_script(session)
            if current_script is not None:
                await session.delete(current_script)

    async def _on_initialize_run_with_statement(self, statement: str) -> None:
        async with self._db.session.begin() as session:
            current_script = await self._load_current_script(session)
            if current_script is not None:
                if current_script.script.script == statement:
                    script = current_script.script
                else:
                    script = Script(script=statement)
                    current_script.script = script
            else:
                script = Script(script=statement)
                current_script = CurrentScript(script=script)
                session.add(current_script)

    async def _load_current_script(self, session: AsyncSession) -> CurrentScript | None:
        stmt = select(CurrentScript).options(selectinload(CurrentScript.script))
        return (await session.execute(stmt)).scalar_one_or_none()



from typing import TYPE_CHECKING, Optional

from sqlalchemy import CheckConstraint, ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_run import Run


class Script(Model):
    __tablename__ = 'script'
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    script: Mapped[str] = mapped_column(Text)

    runs: Mapped[list['Run']] = relationship(back_populates='script')

    _current: Mapped[Optional['CurrentScript']] = relationship(
        'CurrentScript', back_populates='script', cascade='all, delete-orphan'
    )

    @property
    def current(self) -> bool:
        '''True if this script is the current script.

        At most one script can be the current script at any time.
        '''
        return self._current is not None

    @current.setter
    def current(self, value: bool):
        if value and self._current is None:
            self._current = CurrentScript(script=self)
        elif not value:
            self._current = None


class CurrentScript(Model):
    '''The table that keeps track of the current script.

    This table is a singleton, i.e., it can have at most one row. This is
    enforced by a check constraint ensuring the 'id' is always 1.

    '''

    __tablename__ = 'current_script'
    id: Mapped[int] = mapped_column(primary_key=True, index=True, default=1)
    script_id: Mapped[int] = mapped_column(ForeignKey('script.id'))
    script: Mapped[Script] = relationship('Script', back_populates='_current')

    __table_args__ = (CheckConstraint('id = 1'),)



__all__ = [
    'mark_last',
    'all_declared_models_based_on',
    'class_name_and_primary_keys_of',
    'load_all',
    'primary_keys_of',
    'safe_compare',
    'safe_max',
    'safe_min',
    'ensure_async_url',
    'ensure_sync_url',
    'is_timezone_aware',
    'utc_timestamp',
]

from .gen import mark_last
from .sa import (
    all_declared_models_based_on,
    class_name_and_primary_keys_of,
    load_all,
    primary_keys_of,
)
from .safe import safe_compare, safe_max, safe_min
from .url import ensure_async_url, ensure_sync_url
from .utc import is_timezone_aware, utc_timestamp



from typing import Optional

from hypothesis import strategies as st

from nextline_rdb.utils.strategies import st_datetimes

from .. import Run, Stdout, Trace
from .st_trace import st_model_trace


@st.composite
def st_model_stdout(draw: st.DrawFn, trace: Optional[Trace] = None) -> Stdout:
    trace = trace or draw(st_model_trace())
    run_no = trace.run_no
    trace_no = trace.trace_no
    text = draw(st.text())
    written_at = draw(st_datetimes())
    stdout = Stdout(
        run_no=run_no,
        trace_no=trace_no,
        text=text,
        written_at=written_at,
        run=trace.run,
        trace=trace,
    )
    return stdout


@st.composite
def st_model_stdout_list(
    draw: st.DrawFn,
    run: Optional[Run] = None,
    min_size: int = 0,
    max_size: Optional[int] = None,
) -> list[Stdout]:
    run = run or draw(st_model_trace()).run

    traces = (
        draw(
            st.lists(st.sampled_from(run.traces), min_size=min_size, max_size=max_size)
        )
        if run.traces
        else []
    )

    stdouts = list[Stdout]()
    for trace in traces:
        stdout = draw(st_model_stdout(trace=trace))
        stdouts.append(stdout)

    return stdouts



from collections.abc import Iterable
from typing import Callable, Optional, cast

from hypothesis import strategies as st

from nextline_rdb.utils.strategies import st_graphql_ints, st_none_or

from .. import Run, Trace
from .st_run import st_model_run
from .utils import st_started_at_ended_at


@st.composite
def st_model_trace(
    draw: st.DrawFn,
    run: Optional[Run] = None,
    trace_no: Optional[int] = None,
    thread_task_no: Optional[tuple[int, int | None]] = None,
    generate_prompts: bool = False,
) -> Trace:
    from .st_prompt import st_model_prompt_list

    if run is not None:
        # Unable to meet the unique constraint on the prompts table
        assert not generate_prompts

    if trace_no is None:
        trace_no = draw(st_graphql_ints(min_value=1))

    state = draw(st.text())

    thread_task_no = thread_task_no or draw(st_thread_task_no())
    thread_no, task_no = thread_task_no

    trace = Trace(
        trace_no=trace_no,
        state=state,
        thread_no=thread_no,
        task_no=task_no,
    )

    if run is None:
        run = draw(st_model_run(generate_traces=False))

    started_at, ended_at = draw(
        st_started_at_ended_at(
            min_start=run.started_at,
            max_end=run.ended_at,
            allow_start_none=False,
        )
    )

    assert started_at

    trace.run_no = run.run_no
    trace.started_at = started_at
    trace.ended_at = ended_at
    trace.run = run

    if generate_prompts:
        draw(st_model_prompt_list(run=run, min_size=1, max_size=10))

    return trace


def st_thread_task_no() -> st.SearchStrategy[tuple[int, int | None]]:
    return st.tuples(
        st_graphql_ints(min_value=1),
        st_none_or(st_graphql_ints(min_value=1)),
    )


def sort_thread_task_nos(
    thread_task_nos: Iterable[tuple[int, int | None]]
) -> list[tuple[int, int | None]]:
    return sorted(thread_task_nos, key=lambda x: (x[0], x[1] or 0))


@st.composite
def st_model_trace_list(
    draw: st.DrawFn,
    run: Optional[Run] = None,
    min_size: int = 0,
    max_size: Optional[int] = None,
) -> list[Trace]:
    thread_task_nos = draw(
        st.lists(
            st_thread_task_no(), min_size=min_size, max_size=max_size, unique=True
        ).map(sort_thread_task_nos)
    )
    trace_nos = draw(
        st.lists(
            st_graphql_ints(min_value=1),
            min_size=len(thread_task_nos),
            max_size=len(thread_task_nos),
            unique=True,
        ).map(cast(Callable[[Iterable[int]], list[int]], sorted))
    )
    assert len(thread_task_nos) == len(trace_nos)
    traces = list[Trace]()
    for trace_no, thread_task_no in zip(trace_nos, thread_task_nos):
        run = run or draw(st_model_run(generate_traces=False))
        trace = draw(
            st_model_trace(
                run=run,
                trace_no=trace_no,
                thread_task_no=thread_task_no,
            )
        )
        assert run is trace.run
        assert trace_no == trace.trace_no
        assert thread_task_no == (trace.thread_no, trace.task_no)
        traces.append(trace)
    return traces



import strawberry
from hypothesis import given, note
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.models import Run
from nextline_rdb.models.strategies import st_model_run_list
from nextline_rdb.schema import Mutation, Query
from nextline_rdb.utils.strategies import st_graphql_ints

from ..graphql import MUTATE_RDB_DELETE_RUNS


@st.composite
def st_query_variables(
    draw: st.DrawFn, runs: list[Run]
) -> tuple[dict[str, list[int]], list[Run], list[Run]]:
    to_keep = list(runs)
    to_delete = list[Run]()
    ids = list[int]()
    while to_keep:
        run = draw(st.sampled_from(to_keep))
        to_keep.remove(run)
        to_delete.append(run)
        ids.append(run.id)
        if draw(st.booleans()):
            break
    extra = draw(
        st.lists(
            st_graphql_ints(min_value=1).filter(
                lambda i: i not in [run.id for run in runs]
            ),
            max_size=3,
            unique=True,
        )
    )
    ids.extend(extra)
    variables = {'ids': ids}
    return variables, to_keep, to_delete


@given(st.data())
async def test_delete_runs(data: st.DataObject) -> None:
    schema = strawberry.Schema(query=Query, mutation=Mutation)

    runs = data.draw(st_model_run_list(generate_traces=False, min_size=0, max_size=5))

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        note(f'runs: {runs}')

        variables, to_keep, to_delete = data.draw(st_query_variables(runs=runs))

        resp = await schema.execute(
            MUTATE_RDB_DELETE_RUNS,
            variable_values=variables,
            context_value={'db': db},
        )
        note(f'resp: {resp}')

        assert resp.data

        expected = {run.id for run in to_delete}
        assert set(resp.data['rdb']['deleteRuns']) == expected

        async with db.session() as session:
            stmt = select(Run)
            runs_ = (await session.scalars(stmt)).all()
            expected = {run.id for run in to_keep}
            assert {run.id for run in runs_} == expected



import datetime
from typing import TYPE_CHECKING, Annotated, Optional

import strawberry

from nextline_rdb import models as db_models

if TYPE_CHECKING:
    from .run_node import RunNode
    from .trace_node import TraceNode


@strawberry.type
class PromptNode:
    _model: strawberry.Private[db_models.Prompt]
    id: int
    run_no: int
    trace_no: int
    prompt_no: int
    open: bool
    event: str
    started_at: datetime.datetime
    file_name: Optional[str] = None
    line_no: Optional[int] = None
    stdout: Optional[str] = None
    command: Optional[str] = None
    ended_at: Optional[datetime.datetime] = None

    @strawberry.field
    def run(self) -> Annotated['RunNode', strawberry.lazy('.run_node')]:
        from .run_node import RunNode

        return RunNode.from_model(self._model.run)

    @strawberry.field
    def trace(self) -> Annotated['TraceNode', strawberry.lazy('.trace_node')]:
        from .trace_node import TraceNode

        return TraceNode.from_model(self._model.trace)

    @classmethod
    def from_model(cls: type['PromptNode'], model: db_models.Prompt):
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            trace_no=model.trace_no,
            prompt_no=model.prompt_no,
            open=model.open,
            event=model.event,
            started_at=model.started_at,
            file_name=model.file_name,
            line_no=model.line_no,
            stdout=model.stdout,
            command=model.command,
            ended_at=model.ended_at,
        )



import pytest
from hypothesis import HealthCheck, given, settings

from nextline_rdb.utils.strategies import st_python_scripts


@given(st_python_scripts())
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_one(capsys: pytest.CaptureFixture, script: str) -> None:
    exec(script, globals())
    capsys.readouterr()  # clear



import datetime
from typing import TYPE_CHECKING, Annotated, Optional, cast

import strawberry
from sqlalchemy import select
from strawberry.types import Info

from nextline_rdb import models as db_models
from nextline_rdb.db import DB
from nextline_rdb.models import Prompt, Stdout
from nextline_rdb.pagination import SortField

from ..pagination import Connection, load_connection

if TYPE_CHECKING:
    from .prompt_node import PromptNode
    from .run_node import RunNode
    from .stdout_node import StdoutNode


async def _resolve_prompts(
    info: Info,
    root: 'TraceNode',
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> 'Connection[PromptNode]':
    from .prompt_node import PromptNode

    sort = [SortField('prompt_no')]
    select_model = select(Prompt).where(Prompt.trace_id == root._model.id)
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Prompt,
            create_node_from_model=PromptNode.from_model,
            select_model=select_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


async def _resolve_stdouts(
    info: Info,
    root: 'TraceNode',
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> 'Connection[StdoutNode]':
    from .stdout_node import StdoutNode

    sort = [SortField('written_at')]
    select_model = select(Stdout).where(Stdout.trace_id == root._model.id)
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Stdout,
            create_node_from_model=StdoutNode.from_model,
            select_model=select_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


@strawberry.type
class TraceNode:
    _model: strawberry.Private[db_models.Trace]
    id: int
    run_no: int
    trace_no: int
    state: str
    thread_no: int
    started_at: datetime.datetime
    task_no: Optional[int]
    ended_at: Optional[datetime.datetime]

    @strawberry.field
    def run(self) -> Annotated['RunNode', strawberry.lazy('.run_node')]:
        from .run_node import RunNode

        return RunNode.from_model(self._model.run)

    prompts: Connection[
        Annotated['PromptNode', strawberry.lazy('.prompt_node')]
    ] = strawberry.field(resolver=_resolve_prompts)

    stdouts: Connection[
        Annotated['StdoutNode', strawberry.lazy('.stdout_node')]
    ] = strawberry.field(resolver=_resolve_stdouts)

    @classmethod
    def from_model(cls: type['TraceNode'], model: db_models.Trace):
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            trace_no=model.trace_no,
            state=model.state,
            thread_no=model.thread_no,
            started_at=model.started_at,
            task_no=model.task_no,
            ended_at=model.ended_at,
        )



from typing import Optional, cast

import strawberry
from sqlalchemy import select
from sqlalchemy.orm import selectinload
from strawberry.types import Info

import nextline_rdb
from nextline_rdb.db import DB
from nextline_rdb.models import Prompt, Run, Stdout, Trace
from nextline_rdb.pagination import SortField

from .nodes import PromptNode, RunNode, StdoutNode, TraceNode
from .pagination import Connection, load_connection


async def resolve_run(
    info: Info, id: Optional[int] = None, run_no: Optional[int] = None
) -> RunNode | None:
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        # stmt = select(Run).options(selectinload('*'))
        stmt = select(Run).options(selectinload(Run.script))
        if id is not None:
            stmt = stmt.filter(Run.id == id)
        else:
            stmt = stmt.filter(Run.run_no == run_no)
        run = (await session.execute(stmt)).scalar_one_or_none()
    return RunNode.from_model(run) if run else None


async def resolve_runs(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[RunNode]:
    sort = [SortField('run_no', desc=True)]
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Run,
            create_node_from_model=RunNode.from_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


async def resolve_traces(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[TraceNode]:
    sort = [SortField('run_no'), SortField('trace_no')]
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Trace,
            create_node_from_model=TraceNode.from_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


async def resolve_prompts(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[PromptNode]:
    sort = [SortField('run_no'), SortField('prompt_no')]
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Prompt,
            create_node_from_model=PromptNode.from_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


async def resolve_stdouts(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[StdoutNode]:
    sort = [SortField('run_no'), SortField('id')]
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Stdout,
            create_node_from_model=StdoutNode.from_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


def resolve_migration_version(info: Info) -> str | None:
    db = cast(DB, info.context['db'])
    return db.migration_revision


@strawberry.type
class QueryRDB:
    runs: Connection[RunNode] = strawberry.field(resolver=resolve_runs)
    traces: Connection[TraceNode] = strawberry.field(resolver=resolve_traces)
    prompts: Connection[PromptNode] = strawberry.field(resolver=resolve_prompts)
    stdouts: Connection[StdoutNode] = strawberry.field(resolver=resolve_stdouts)
    run: RunNode | None = strawberry.field(resolver=resolve_run)
    version: str = nextline_rdb.__version__
    migration_version: str | None = strawberry.field(resolver=resolve_migration_version)


@strawberry.type
class Query:
    @strawberry.field
    async def history(self) -> QueryRDB:
        # TODO: Remove this method.
        return QueryRDB()

    @strawberry.field
    async def rdb(self) -> QueryRDB:
        return QueryRDB()



from collections.abc import Iterable
from typing import Callable, Optional, cast

from hypothesis import strategies as st

from nextline_rdb.utils.strategies import st_graphql_ints, st_none_or

from .. import Prompt, Run, Trace
from .st_trace import st_model_trace
from .utils import st_started_at_ended_at


@st.composite
def st_model_prompt(
    draw: st.DrawFn,
    prompt_no: Optional[int] = None,
    trace: Optional[Trace] = None,
) -> Prompt:
    trace = trace or draw(st_model_trace())
    run_no = trace.run_no
    trace_no = trace.trace_no
    if prompt_no is None:
        prompt_no = draw(st_graphql_ints(min_value=1))
    open = draw(st.booleans())
    event = draw(st.text())
    started_at, ended_at = draw(
        st_started_at_ended_at(
            min_start=trace.started_at,
            max_end=trace.ended_at,
            allow_start_none=False,
        )
    )
    file_name = draw(st_none_or(st.text()))
    line_no = draw(st_none_or(st_graphql_ints(min_value=1)))
    stdout = draw(st_none_or(st.text()))
    command = draw(st_none_or(st.text()))
    model = Prompt(
        run_no=run_no,
        trace_no=trace_no,
        prompt_no=prompt_no,
        open=open,
        event=event,
        started_at=started_at,
        file_name=file_name,
        line_no=line_no,
        stdout=stdout,
        command=command,
        ended_at=ended_at,
        run=trace.run,
        trace=trace,
    )
    return model


@st.composite
def st_model_prompt_list(
    draw: st.DrawFn,
    run: Optional[Run] = None,
    min_size: int = 0,
    max_size: Optional[int] = None,
) -> list[Prompt]:
    # NOTE: Unique constraint: (run_no, prompt_no)
    run = run or draw(st_model_trace()).run

    if not run.traces:
        return []

    prompt_nos = draw(
        st.lists(
            st_graphql_ints(min_value=1),
            min_size=min_size,
            max_size=max_size,
            unique=True,
        ).map(cast(Callable[[Iterable[int]], list[int]], sorted))
    )

    size = len(prompt_nos)

    traces = draw(st.lists(st.sampled_from(run.traces), min_size=size, max_size=size))

    prompts = list[Prompt]()
    for trace, prompt_no in zip(traces, prompt_nos):
        prompt = draw(st_model_prompt(prompt_no=prompt_no, trace=trace))
        prompts.append(prompt)
    return prompts



from collections.abc import Sequence
from typing import Any, NamedTuple, Optional, Type, TypeVar

from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase, aliased, selectinload
from sqlalchemy.sql.selectable import Select

# import sqlparse


# def format_sql(sql):
#     return sqlparse.format(str(sql), reindent=True, keyword_case='upper')


class SortField(NamedTuple):
    field: str
    desc: bool = False


Sort = list[SortField]

_Id = TypeVar('_Id')

T = TypeVar('T', bound=DeclarativeBase)


async def load_models(
    session: AsyncSession,
    Model: Type[T],
    id_field: str,
    *,
    select_model: Optional[Select[tuple[T]]] = None,
    sort: Optional[Sort] = None,
    before: Optional[_Id] = None,
    after: Optional[_Id] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
):
    sort = sort or []

    if id_field not in {s.field for s in sort}:
        sort.append(SortField(id_field))

    order_by = [
        f.desc() if d else f
        for f, d in [(getattr(Model, s.field), s.desc) for s in sort]
    ]

    stmt = compose_statement(
        Model,
        id_field,
        select_model=select_model,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
    )

    # TODO: Make this optional
    stmt = stmt.options(selectinload('*'))

    models = await session.scalars(stmt)
    return models


def compose_statement(
    Model: Type[T],
    id_field: str,
    *,
    select_model: Optional[Select[tuple[T]]] = None,
    order_by: Optional[Sequence[Any]] = None,
    before: Optional[_Id] = None,
    after: Optional[_Id] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Select[tuple[T]]:
    '''Return a SQL select statement for pagination.

    Parameters
    ----------
    Model :
        The class of the ORM model to query.
    id_field :
        The name of the primary key field, e.g., 'id'.
    select_model : optional
        E.g., `select(Model).where(...)`. If not provided, `select(Model)` is used.
    order_by : optional
        The arguments to `row_number().over(order_by=...)` and `order_by()`. If
        not provided, the primary key field is used, e.g., `[Model.id]`.
    before : optional
        As in the GraphQL Cursor Connections Specification [1].
    after : optional
        As in the GraphQL Cursor Connections Specification [1].
    first : optional
        As in the GraphQL Cursor Connections Specification [1].
    last : optional
        As in the GraphQL Cursor Connections Specification [1].

    Returns
    -------
    stmt
        The composed select statement for pagination.

    Raises
    ------
    ValueError
        If both before/last and after/first parameters are provided.

    References
    ----------
    .. [1] https://relay.dev/graphql/connections.htm
    '''

    forward = (after is not None) or (first is not None)
    backward = (before is not None) or (last is not None)

    if forward and backward:
        raise ValueError('Only either after/first or before/last is allowed')

    if select_model is None:
        select_model = select(Model)

    if not order_by:
        # E.g., [T.id]
        order_by = [getattr(Model, id_field)]

    if not (forward or backward):
        return select_model.order_by(*order_by)

    cursor = after if forward else before
    limit = first if forward else last

    # A CTE (Common Table Expression) with a row_number column
    cte = select_model.add_columns(
        func.row_number().over(order_by=order_by).label('row_number')
    ).cte()

    Alias = aliased(Model, cte)
    stmt = select(Alias, cte.c.row_number).select_from(cte)

    if cursor is not None:
        # A subquery to find the row_number at the cursor
        stmt_subq = select(cte.c.row_number.label('at_cursor'))
        stmt_subq = stmt_subq.where(getattr(cte.c, id_field) == cursor)
        subq = stmt_subq.subquery()

        # Select rows after or before (if backward) the cursor
        stmt = stmt.select_from(subq)
        if backward:
            stmt = stmt.where(cte.c.row_number < subq.c.at_cursor)
        else:
            stmt = stmt.where(cte.c.row_number > subq.c.at_cursor)

    if limit is not None:
        # Specify the maximum number of rows to return
        if backward:
            stmt = stmt.order_by(cte.c.row_number.desc())
        else:
            stmt = stmt.order_by(cte.c.row_number)
        stmt = stmt.limit(limit)

    # Select only the model (not the row_number) and ensure the order
    cte = stmt.cte()
    Alias = aliased(Model, cte)
    stmt = select(Alias).select_from(cte)
    stmt = stmt.order_by(cte.c.row_number)

    return stmt



from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.db import DB
from nextline_rdb.utils import class_name_and_primary_keys_of, load_all
from nextline_rdb.utils.strategies import st_ranges

from ... import Model
from .. import st_model_instance_list


@given(st.data())
async def test_options(data: st.DataObject) -> None:
    # Generate options of the strategy to be tested
    min_size, max_size = data.draw(
        st_ranges(
            st_=st.integers,
            min_start=0,
            max_end=8,
            allow_start_none=False,
            allow_end_none=False,
        )
    )
    assert isinstance(min_size, int)
    assert isinstance(max_size, int)

    # Call the strategy to be tested
    instances = data.draw(st_model_instance_list(min_size=min_size, max_size=max_size))

    # Assert the generated values
    assert min_size <= len(instances) <= max_size


@given(instances=st_model_instance_list(min_size=0, max_size=5))
async def test_db(instances: list[Model]) -> None:
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add_all(instances)
            added = list(session.new)
            assert set(instances) <= set(added)
        added = sorted(added, key=class_name_and_primary_keys_of)
        repr_added = [repr(m) for m in added]

        async with db.session() as session:
            loaded = await load_all(session, Model)
            repr_loaded = [repr(m) for m in loaded]

    assert repr_added == repr_loaded



import asyncio
from collections.abc import AsyncIterator, Callable
from pathlib import Path
from typing import Optional

import pytest
from nextline import Nextline
from nextline.events import OnStartPrompt
from nextline.plugin.spec import Context, hookimpl
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb import DB
from nextline_rdb import models as db_models
from nextline_rdb import write


async def test_one(adb: DB, run_nextline, statement):
    del run_nextline

    async with adb.session() as session:
        stmt = select(db_models.Run).options(selectinload(db_models.Run.script))
        runs = (await session.scalars(stmt)).all()
        assert 1 == len(runs)
        run = runs[0]
        assert 2 == run.run_no
        assert run.started_at
        assert run.ended_at
        assert run.script
        assert statement == run.script.script
        assert not run.exception

        traces = (await session.scalars(select(db_models.Trace))).all()
        assert 5 == len(traces)
        run_no = 2
        trace_no = 0
        for trace in traces:
            trace_no += 1
            assert trace_no == trace.trace_no
            assert run_no == trace.run_no
            assert trace.started_at
            assert trace.ended_at

        prompts = (await session.scalars(select(db_models.Prompt))).all()
        assert 58 == len(prompts)
        for prompt in prompts:
            assert run_no == prompt.run_no
            assert prompt.trace_no
            assert prompt.started_at
            assert prompt.line_no
            assert prompt.file_name
            assert prompt.event

        stdouts = (await session.scalars(select(db_models.Stdout))).all()
        assert 1 == len(stdouts)
        # for stdout in stdouts:
        #     assert run_no == stdout.run_no
        #     assert stdout.text
        #     assert stdout.written_at


@pytest.fixture
def monkey_patch_syspath(monkeypatch: pytest.MonkeyPatch) -> None:
    here = Path(__file__).resolve().parent
    path = here / 'example_script'
    monkeypatch.syspath_prepend(str(path))
    return


@pytest.fixture
def statement(monkey_patch_syspath: None) -> str:
    del monkey_patch_syspath
    here = Path(__file__).resolve().parent
    path = here / 'example_script'
    return (path / 'script.py').read_text()


@pytest.fixture
async def run_nextline(adb: DB, statement: str) -> None:
    nextline = Nextline(statement, trace_threads=True, trace_modules=True)
    write.register(nextline=nextline, db=adb)
    nextline.register(Control())
    async with nextline:
        await run_statement(nextline, statement)


@pytest.fixture
async def adb(url: str) -> AsyncIterator[DB]:
    async with DB(url=url) as adb:
        yield adb


@pytest.fixture
def url(tmp_url_factory: Callable[[], str]) -> str:
    return tmp_url_factory()


@pytest.fixture(scope='session')
def tmp_url_factory(tmp_path_factory: pytest.TempPathFactory) -> Callable[[], str]:
    def factory() -> str:
        dir = tmp_path_factory.mktemp('db')
        url = f'sqlite:///{dir}/db.sqlite'
        return url

    return factory


async def run_statement(nextline: Nextline, statement: Optional[str] = None) -> None:
    await asyncio.sleep(0.01)
    await nextline.reset(statement=statement)
    await asyncio.sleep(0.01)
    async with nextline.run_session():
        pass
    await asyncio.sleep(0.01)


class Control:
    @hookimpl
    async def on_start_prompt(self, context: Context, event: OnStartPrompt) -> None:
        nextline = context.nextline
        if event.event == 'line':
            line = nextline.get_source_line(
                line_no=event.line_no, file_name=event.file_name
            )
            command = find_command(line) or 'next'
        else:
            command = 'next'
        await asyncio.sleep(0.01)
        await nextline.send_pdb_command(command, event.prompt_no, event.trace_no)


def find_command(line: str) -> Optional[str]:
    '''The Pdb command indicated in a comment

    For example, returns 'step' for the line 'func()  # step'
    '''
    import re

    if not (comment := extract_comment(line)):
        return None
    regex = re.compile(r'^# +(\w+) *$')
    match = regex.search(comment)
    if match:
        return match.group(1)
    return None


def extract_comment(line: str) -> Optional[str]:
    import io
    import tokenize

    comments = [
        val
        for type, val, *_ in tokenize.generate_tokens(io.StringIO(line).readline)
        if type == tokenize.COMMENT
    ]
    if comments:
        return comments[0]
    return None



from typing import TYPE_CHECKING

from sqlalchemy import CheckConstraint, ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_run import Run


class Script(Model):
    __tablename__ = 'script'
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    current: Mapped[bool] = mapped_column(default=False)
    script: Mapped[str] = mapped_column(Text)

    runs: Mapped[list['Run']] = relationship(back_populates='script')

    _current: Mapped['CurrentScript'] = relationship(
        'CurrentScript', back_populates='script', cascade='all, delete-orphan'
    )


class CurrentScript(Model):
    '''The table that keeps track of the current script.

    This table is a singleton, i.e., it can have at most one row. This is
    enforced by a check constraint ensuring the 'id' is always 1.

    '''

    __tablename__ = 'current_script'
    id: Mapped[int] = mapped_column(primary_key=True, index=True, default=1)
    script_id: Mapped[int] = mapped_column(ForeignKey('script.id'))
    script: Mapped[Script] = relationship('Script', back_populates='_current')

    __table_args__ = (CheckConstraint('id = 1'),)



import enum
import math
from typing import Any

from sqlalchemy.orm import ColumnProperty, class_mapper


class ReprMixin:
    def __repr__(self) -> str:
        '''Return a string that can be used to reconstruct the object.'''
        props = class_mapper(self.__class__).iterate_properties
        columns = (p for p in props if isinstance(p, ColumnProperty))
        keys = (c.key for c in columns if self._repr_filter(c.key))
        fmts = (f'{k}={self._repr_val(k)}' for k in keys)
        return f'{self.__class__.__qualname__}({", ".join(fmts)})'

    def _repr_filter(self, key: str) -> bool:
        '''Exclude if the value is None. Override this method if necessary.'''
        val = getattr(self, key)
        return val is not None

    def _repr_val(self, key: str) -> str:
        '''Format the value so it can be reconstructed. Override this method if necessary.'''
        val = getattr(self, key)
        return repr_val(val)


def repr_val(val: Any) -> str:
    match val:
        case enum.Enum():
            return str(val)
        case int():
            return f'{val:_}'
        case math.inf:
            return 'float("inf")'
        case float(neg_inf) if neg_inf == float('-inf'):
            return 'float("-inf")'
        case float(nan_val) if math.isnan(nan_val):
            return 'float("nan")'
        case _:
            return repr(val)



import datetime
from typing import TYPE_CHECKING, Annotated, Optional, cast

import strawberry
from sqlalchemy import select
from strawberry.types import Info

from nextline_rdb.db import DB
from nextline_rdb.models import Prompt, Run, Stdout, Trace
from nextline_rdb.pagination import SortField

from ..pagination import Connection, load_connection

if TYPE_CHECKING:
    from .prompt_node import PromptNode
    from .stdout_node import StdoutNode
    from .trace_node import TraceNode


async def _resolve_traces(
    info: Info,
    root: 'RunNode',
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection['TraceNode']:
    from .trace_node import TraceNode

    sort = [SortField('trace_no')]
    select_model = select(Trace).where(Trace.run_id == root._model.id)
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Trace,
            create_node_from_model=TraceNode.from_model,
            select_model=select_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


async def _resolve_prompts(
    info: Info,
    root: 'RunNode',
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection['PromptNode']:
    from .prompt_node import PromptNode

    sort = [SortField('prompt_no')]
    select_model = select(Prompt).where(Prompt.run_id == root._model.id)
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Prompt,
            create_node_from_model=PromptNode.from_model,
            select_model=select_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


async def _resolve_stdouts(
    info: Info,
    root: 'RunNode',
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection['StdoutNode']:
    from .stdout_node import StdoutNode

    sort = [SortField('written_at')]
    select_model = select(Stdout).where(Stdout.run_id == root._model.id)
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        return await load_connection(
            session,
            Stdout,
            create_node_from_model=StdoutNode.from_model,
            select_model=select_model,
            sort=sort,
            before=before,
            after=after,
            first=first,
            last=last,
        )


@strawberry.type
class RunNode:
    _model: strawberry.Private[Run]
    id: int
    run_no: int
    state: Optional[str]
    started_at: Optional[datetime.datetime]
    ended_at: Optional[datetime.datetime]
    script: Optional[str]
    exception: Optional[str]

    traces: Connection[
        Annotated['TraceNode', strawberry.lazy('.trace_node')]
    ] = strawberry.field(resolver=_resolve_traces)

    prompts: Connection[
        Annotated['PromptNode', strawberry.lazy('.prompt_node')]
    ] = strawberry.field(resolver=_resolve_prompts)

    stdouts: Connection[
        Annotated['StdoutNode', strawberry.lazy('.stdout_node')]
    ] = strawberry.field(resolver=_resolve_stdouts)

    @classmethod
    def from_model(cls: type['RunNode'], model: Run):
        script = model.script.script if model.script else None
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            state=model.state,
            started_at=model.started_at,
            ended_at=model.ended_at,
            script=script,
            exception=model.exception,
        )
